@@ -22,8 +22,10 @@ import json
 import logging
 import os
 import platform
+import shutil
 import subprocess
 import sys
+import time as pytime
 import traceback
 import uuid
 
@@ -84,6 +86,7 @@ from guake.utils import HidePrevention
 from guake.utils import RectCalculator
 from guake.utils import TabNameUtils
 from guake.utils import get_server_time
+from guake.utils import save_tabs_when_changed
 
 from locale import gettext as _
 
@@ -133,8 +136,6 @@ class Guake(SimpleGladeApp):
         patch_gtk_theme(self.get_widget("window-root").get_style_context(), self.settings)
         self.add_callbacks(self)
 
-        self.debug_mode = self.settings.general.get_boolean('debug-mode')
-        setupLogging(self.debug_mode)
         log.info('Guake Terminal %s', guake_version())
         log.info('VTE %s', vte_version())
         log.info('Gtk %s', gtk_version())
@@ -211,6 +212,9 @@ class Guake(SimpleGladeApp):
 
         self.window.connect('draw', draw_callback)
 
+        # Debounce accel_search_terminal
+        self.prev_accel_search_terminal_time = 0.0
+
         # holds the timestamp of the losefocus event
         self.losefocus_time = 0
 
@@ -263,6 +267,10 @@ class Guake(SimpleGladeApp):
 
         refresh_user_start(self.settings)
 
+        # Restore tabs when startup
+        if self.settings.general.get_boolean('restore-tabs-startup'):
+            self.restore_tabs(suppress_notify=True)
+
         # Pop-up that shows that guake is working properly (if not
         # unset in the preferences windows)
         if self.settings.general.get_boolean('use-popup'):
@@ -284,6 +292,9 @@ class Guake(SimpleGladeApp):
     def notebook_created(self, nm, notebook, key):
         notebook.attach_guake(self)
 
+        # Tracking when reorder page
+        notebook.connect('page-reordered', self.on_page_reorder)
+
     # new color methods should be moved to the GuakeTerminal class
 
     def _load_palette(self):
@@ -704,6 +715,27 @@ class Guake(SimpleGladeApp):
         self.settings.general.triggerOnChangedValue(self.settings.general, 'compat-backspace')
         self.settings.general.triggerOnChangedValue(self.settings.general, 'compat-delete')
 
+    def accel_search_terminal(self, *args):
+        nb = self.get_notebook()
+        term = nb.get_current_terminal()
+        box = nb.get_nth_page(nb.find_page_index_by_terminal(term))
+
+        # Debounce it
+        current_time = pytime.time()
+        if current_time - self.prev_accel_search_terminal_time < 0.3:
+            return
+
+        self.prev_accel_search_terminal_time = current_time
+        if box.search_revealer.get_reveal_child():
+            if box.search_entry.has_focus():
+                box.hide_search_box()
+            else:
+                # The box was showed, but out of focus
+                # Don't hide it, re-grab the focus to search entry
+                box.search_entry.grab_focus()
+        else:
+            box.show_search_box()
+
     def accel_quit(self, *args):
         """Callback to prompt the user whether to quit Guake or not.
         """
@@ -926,21 +958,29 @@ class Guake(SimpleGladeApp):
             pass
         return TabNameUtils.shorten(vte_title, self.settings)
 
+    @save_tabs_when_changed
     def on_terminal_title_changed(self, vte, term):
         # box must be a page
         box = term.get_parent().get_root_box()
         use_vte_titles = self.settings.general.get_boolean('use-vte-titles')
         if not use_vte_titles:
             return
+
+        # NOTE: Try our best to find the page_num inside all notebooks
         # this may return -1, should be checked ;)
-        page_num = self.get_notebook().page_num(box)
+        nb = self.get_notebook()
+        page_num = nb.page_num(box)
+        for nb in self.notebook_manager.iter_notebooks():
+            page_num = nb.page_num(box)
+            if page_num != -1:
+                break
         # if tab has been renamed by user, don't override.
         if not getattr(box, 'custom_label_set', False):
             title = self.compute_tab_title(vte)
-            self.get_notebook().rename_page(page_num, title, False)
+            nb.rename_page(page_num, title, False)
             self.update_window_title(title)
         else:
-            text = self.get_notebook().get_tab_text_page(box)
+            text = nb.get_tab_text_page(box)
             if text:
                 self.update_window_title(text)
 
@@ -978,6 +1018,7 @@ class Guake(SimpleGladeApp):
         self.load_config()
         terminal.connect('window-title-changed', self.on_terminal_title_changed, terminal)
 
+    @save_tabs_when_changed
     def add_tab(self, directory=None):
         """Adds a new tab to the terminal notebook.
         """
@@ -1099,3 +1140,107 @@ class Guake(SimpleGladeApp):
                 log.debug(traceback.format_exc())
             else:
                 log.debug("hook on event %s has been executed", event_name)
+
+    @save_tabs_when_changed
+    def on_page_reorder(self, notebook, child, page_num):
+        # Yep, just used for save tabs when changed
+        pass
+
+    def get_xdg_config_directory(self):
+        return Path('~/.config/guake').expanduser()
+
+    def save_tabs(self, filename='session.json'):
+        config = {'schema_version': 1, 'timestamp': int(pytime.time()), 'workspace': {}}
+        for key, nb in self.notebook_manager.get_notebooks().items():
+            tabs = []
+            for index in range(nb.get_n_pages()):
+                try:
+                    page = nb.get_nth_page(index)
+                    if page.child:
+                        tabs.append({
+                            'directory': page.child.terminal.get_current_directory(),
+                            'label': nb.get_tab_text_index(index),
+                            'custom_label_set': getattr(page, 'custom_label_set', False)
+                        })
+                except FileNotFoundError:
+                    # discard same broken tabs
+                    pass
+            # NOTE: Maybe we will have frame inside the workspace in future
+            #       So lets use list to store the tabs (as for each frame)
+            config['workspace'][key] = [tabs]
+
+        if not self.get_xdg_config_directory().exists():
+            self.get_xdg_config_directory().mkdir(parents=True)
+        session_file = self.get_xdg_config_directory() / filename
+        with session_file.open('w') as f:
+            json.dump(config, f, ensure_ascii=False, indent=4)
+        log.info('Guake tabs saved to %s', session_file)
+
+    def restore_tabs(self, filename='session.json', suppress_notify=False):
+        session_file = self.get_xdg_config_directory() / filename
+        if not session_file.exists():
+            log.info('Cannot found tabs session.json file')
+            return
+        with session_file.open() as f:
+            try:
+                config = json.load(f)
+            except Exception:
+                log.warning('%s is broken', session_file)
+                shutil.copy(
+                    session_file,
+                    self.get_xdg_config_directory() / '{0}.bak'.format(filename)
+                )
+                img_filename = pixmapfile('guake-notification.png')
+                notifier.showMessage(
+                    _('Guake Terminal'),
+                    _('Your {session_filename} file is broken, backup to {session_filename}.bak'
+                      ).format(session_filename=filename), img_filename
+                )
+                return
+
+        # Disable auto save tabs
+        v = self.settings.general.get_boolean('save-tabs-when-changed')
+        self.settings.general.set_boolean('save-tabs-when-changed', False)
+
+        # Restore all tabs for all workspaces
+        try:
+            for key, frames in config['workspace'].items():
+                nb = self.notebook_manager.get_notebook(int(key))
+                current_pages = nb.get_n_pages()
+
+                # Restore each frames' tabs from config
+                # NOTE: If frame implement in future, we will need to update this code
+                for tabs in frames:
+                    for index, tab in enumerate(tabs):
+                        nb.new_page_with_focus(
+                            tab['directory'], tab['label'], tab['custom_label_set']
+                        )
+
+                    # Remove original pages in notebook
+                    for i in range(current_pages):
+                        nb.delete_page(0)
+        except KeyError:
+            log.warning('%s schema is broken', session_file)
+            shutil.copy(session_file, self.get_xdg_config_directory() / '{}.bak'.format(filename))
+            with (self.get_xdg_config_directory() / '{}.log.err'.format(filename)).open('w') as f:
+                traceback.print_exc(file=f)
+            img_filename = pixmapfile('guake-notification.png')
+            notifier.showMessage(
+                _('Guake Terminal'),
+                _(
+                    'Your {session_filename} schema is broken, backup to {session_filename}.bak, '
+                    'and error message has been saved to {session_filename}.log.err.'.format(
+                        session_filename=filename
+                    )
+                ), img_filename
+            )
+
+        # Reset auto save tabs
+        self.settings.general.set_boolean('save-tabs-when-changed', v)
+
+        # Notify the user
+        if (self.settings.general.get_boolean('restore-tabs-notify') and not suppress_notify):
+            filename = pixmapfile('guake-notification.png')
+            notifier.showMessage(_("Guake Terminal"), _("Your tabs has been restored!"), filename)
+
+        log.info('Guake tabs restored from %s', session_file)
